import threading
from os import remove
from gettext import gettext as _
from gi.repository import Gtk
from gfeeds.confManager import ConfManager
from gfeeds.download_manager import download_raw
from gfeeds.feeds_manager import FeedsManager


class RowPopover(Gtk.Popover):
    def __init__(self, parent, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/article_right_click_popover_content.glade'
        )
        self.container_box = self.builder.get_object('container_box')
        # self.set_size_request(270, 150)
        self.parent = parent

        # self.title_label = self.builder.get_object('title_label')
        # self.title_label.set_text(self.parent.feeditem.title)

        self.read_unread_btn = self.builder.get_object('read_unread_btn')
        self.read_unread_btn.connect('clicked', self.on_read_unread_clicked)
        self.read_unread_img = self.builder.get_object('read_unread_img')
        self.read_unread_label = self.builder.get_object('read_unread_label')

        self.save_btn = self.builder.get_object('save_btn')
        self.save_btn.set_active(
            self.parent.feeditem.link in self.confman.conf['saved_items']
        )
        self.save_btn_handler_id = self.save_btn.connect(
            'toggled', self.on_save_toggled
        )

        self.set_modal(True)
        self.set_relative_to(self.parent)
        self.add(self.container_box)
        if self.parent.feeditem.read:
            self.read_unread_img.set_from_icon_name(
                'eye-not-looking-symbolic',
                Gtk.IconSize.BUTTON
            )
            self.read_unread_label.set_text(_(
                'Mark as unread'
            ))
        else:
            self.read_unread_img.set_from_icon_name(
                'eye-open-negative-filled-symbolic',
                Gtk.IconSize.BUTTON
            )
            self.read_unread_label.set_text(_(
                'Mark as read'
            ))

    def set_read(self, read):
        parent_stack = self.parent.get_parent().get_parent(
        ).get_parent().get_parent()
        other_list = (
            parent_stack.listbox
            if self.parent.get_parent() == parent_stack.saved_items_listbox
            else parent_stack.saved_items_listbox
        )
        other_row = None
        for row in other_list.get_children():
            if row.feeditem.link == self.parent.feeditem.link:
                other_row = row
                break
        rows = [self.parent, ]
        if other_row:
            rows.append(other_row)
        if not read:
            for r in rows:
                r.set_read(False)
                r.popover.read_unread_img.set_from_icon_name(
                    'eye-open-negative-filled-symbolic',
                    Gtk.IconSize.BUTTON
                )
                r.popover.read_unread_label.set_text(_(
                    'Mark as read'
                ))
        else:
            for r in rows:
                r.set_read(True)
                r.popover.read_unread_img.set_from_icon_name(
                    'eye-not-looking-symbolic',
                    Gtk.IconSize.BUTTON
                )
                r.popover.read_unread_label.set_text(_(
                    'Mark as unread'
                ))
        parent_stack.listbox.invalidate_filter()
        other_list.invalidate_filter()

    def on_read_unread_clicked(self, btn):
        self.popdown()
        self.set_read(not self.parent.feeditem.read)

    def on_save_toggled(self, togglebtn):
        self.popdown()
        togglebtn.set_sensitive(False)
        if togglebtn.get_active():
            fi_dict = self.parent.feeditem.to_dict()
            t = threading.Thread(
                group=None,
                target=download_raw,
                name=None,
                args=(
                    fi_dict['link'],
                    self.confman.saved_cache_path + '/' + fi_dict['linkhash']
                )
            )
            t.start()
            self.confman.conf[
                'saved_items'
            ][self.parent.feeditem.link] = fi_dict
        else:
            todel_fi_dict = self.confman.conf['saved_items'].pop(
                self.parent.feeditem.link
            )
            remove(
                self.confman.saved_cache_path + '/' + todel_fi_dict['linkhash']
            )
            # parent.is_saved means "the row that spawned the popover is from
            # the *Saved* section"
            if self.parent.is_saved:
                parent_stack = self.parent.get_parent().get_parent(
                ).get_parent().get_parent()
                parent_stack.on_saved_item_deleted(todel_fi_dict['link'])
        self.confman.save_conf()
        self.feedman.populate_saved_feeds_items()
        togglebtn.set_sensitive(True)



from gi.repository import Gtk


def separator_header_func(row, prev_row=None):
    if (
            prev_row is not None and
            row.get_header() is None
    ):
        row.set_header(
            Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        )



from gi.repository import Gtk, Gdk, GLib
from gettext import gettext as _
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.sidebar_row import GFeedsSidebarRow
from gfeeds.listbox_tools import separator_header_func
from gfeeds.rss_parser import FakeFeed


class GFeedsSidebarListBox(Gtk.ListBox):
    def __init__(self, parent_stack, **kwargs):
        super().__init__(**kwargs)
        self.search_terms = ''
        self.confman = ConfManager()
        self.parent_stack = parent_stack

        self.set_sort_from_confman()
        self.confman.connect(
            'gfeeds_new_first_changed',
            self.set_sort_from_confman
        )
        self.selected_feeds = []
        self.set_filter_func(self.gfeeds_sidebar_filter_func, None, False)
        self.confman.connect(
            'gfeeds_filter_changed',
            self.change_filter
        )
        self.confman.connect(
            'gfeeds_show_read_changed',
            self.on_show_read_changed
        )

        # longpress & right click
        self.longpress = Gtk.GestureLongPress.new(self)
        self.longpress.set_propagation_phase(Gtk.PropagationPhase.TARGET)
        self.longpress.set_touch_only(False)
        self.longpress.connect(
            'pressed',
            self.on_right_click,
            self
        )
        self.connect(
            'button-press-event',
            self.on_key_press_event
        )
        self.set_header_func(separator_header_func)

    def add(self, *args, **kwargs):
        super().add(*args, **kwargs)
        self.show_all()

    def on_show_read_changed(self, *args):
        self.invalidate_filter()

    def on_right_click(self, e_or_g, x, y, *args):
        row = self.get_row_at_y(y)
        if row:
            row.popover.popup()

    def on_key_press_event(self, what, event):
        # right click
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            self.on_right_click(event, event.x, event.y)

    def change_filter(self, caller, n_filter):
        if n_filter is None:
            self.selected_feeds = []
        elif isinstance(n_filter, list):
            n_filter = n_filter[0]
            # filter by tag
            self.selected_feeds = [
                f for f in self.confman.conf['feeds'].keys()
                if 'tags' in self.confman.conf['feeds'][f].keys() and
                n_filter in self.confman.conf['feeds'][f]['tags']
            ]
        else:
            self.selected_feeds = [n_filter.rss_link]
        self.invalidate_filter()

    def gfeeds_sidebar_filter_func(self, row, data, notify_destroy):
        toret = False
        if isinstance(row.feeditem.parent_feed, FakeFeed):
            toret = True
        elif len(self.selected_feeds) <= 0:
            toret = True
        else:
            toret = row.feeditem.parent_feed.rss_link in self.selected_feeds
        return row.is_selected() or (
            toret and (
                self.confman.conf['show_read_items'] or
                not row.feeditem.read
            ) and (
                not self.search_terms or
                self.search_terms.lower() in row.feeditem.title.lower()
            )
        )

    def set_sort_from_confman(self, *args):
        if self.confman.conf['new_first']:
            self.set_sort_func(self.gfeeds_sort_new_first_func, None, False)
        else:
            self.set_sort_func(self.gfeeds_sort_old_first_func, None, False)

    def add_new_items(self, feeditems_l):
        self.parent_stack.set_main_visible(True)
        for i in feeditems_l:
            self.add(GFeedsSidebarRow(i))
            self.show_all()

    def empty(self, *args):
        while True:
            row = self.get_row_at_index(0)
            if row:
                self.remove(row)
            else:
                break

    def populate(self, feeditems_l):
        self.empty()
        self.add_new_items(feeditems_l)

    def gfeeds_sort_new_first_func(self, row1, row2, data, notify_destroy):
        return row1.feeditem.pub_date < row2.feeditem.pub_date

    def gfeeds_sort_old_first_func(self, row1, row2, data, notify_destroy):
        return row1.feeditem.pub_date > row2.feeditem.pub_date


class GFeedsSidebarScrolledWin(Gtk.ScrolledWindow):
    def __init__(self, parent_stack, **kwargs):
        super().__init__(**kwargs)
        self.parent_stack = parent_stack
        self.listbox = GFeedsSidebarListBox(self.parent_stack)
        self.empty = self.listbox.empty
        self.populate = self.listbox.populate
        # self.set_size_request(360, 100)

        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.add(self.listbox)

    def select_next_article(self, *args):
        target = None
        current_row = self.listbox.get_selected_row()
        if not current_row:
            target = self.listbox.get_row_at_index(0)
        else:
            target = self.listbox.get_row_at_index(
                current_row.get_index()+1
            )
        if target:
            self.listbox.select_row(target)
            self.listbox.emit('row-activated', target)

    def select_prev_article(self, *args):
        target = None
        current_row = self.listbox.get_selected_row()
        if not current_row:
            return
        target = self.listbox.get_row_at_index(
            current_row.get_index()-1
        )
        if target:
            self.listbox.select_row(target)
            self.listbox.emit('row-activated', target)


class GFeedsSidebar(Gtk.Stack):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feedman = FeedsManager()
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

        self.scrolled_win = GFeedsSidebarScrolledWin(self)
        self.listbox = self.scrolled_win.listbox
        self.empty = self.scrolled_win.listbox.empty
        self.populate = self.scrolled_win.listbox.populate

        self.saved_items_scrolled_win = GFeedsSidebarScrolledWin(self)
        self.saved_items_listbox = self.saved_items_scrolled_win.listbox

        self.add_titled(self.scrolled_win, 'Feed', _('Feed'))
        self.child_set_property(
            self.scrolled_win,
            'icon-name',
            'application-rss+xml-symbolic'
        )
        self.add_titled(self.saved_items_scrolled_win, 'Saved', _('Saved'))
        self.child_set_property(
            self.saved_items_scrolled_win,
            'icon-name',
            'emblem-favorite-symbolic'
        )
        # self.set_size_request(360, 100)

        self.feedman.feeds_items.connect(
            'pop',
            lambda caller, obj: self.on_feeds_items_pop(obj)
        )
        self.feedman.feeds_items.connect(
            'append',
            lambda caller, obj: GLib.idle_add(
                self.on_feeds_items_append,
                obj,
                priority=GLib.PRIORITY_LOW)
        )
        self.feedman.feeds_items.connect(
            'extend',
            lambda caller, obj: GLib.idle_add(
                self.on_feeds_items_extend,
                caller,
                obj,
                priority=GLib.PRIORITY_LOW
            )
        )
        self.feedman.feeds_items.connect(
            'empty',
            lambda *args: self.listbox.empty()
        )
        self.feedman.saved_feeds_items.connect(
            'empty',
            lambda *args: self.saved_items_listbox.empty()
        )
        self.feedman.saved_feeds_items.connect(
            'append',
            lambda caller, obj: self.on_saved_feeds_items_append(obj)
        )

        self.feedman.feeds.connect(
            'pop',
            lambda caller, obj: self.on_feeds_pop(obj)
        )

    def on_feeds_items_extend(self, caller, n_feeds_items_list):
        [
            self.listbox.add(GFeedsSidebarRow(feed_item))
            for feed_item in n_feeds_items_list
        ]

    def on_feeds_pop(self, obj):
        if obj == self.listbox.selected_feed:
            self.listbox.selected_feed = None
            self.listbox.invalidate_filter()

    def set_search(self, search_terms):
        for lb in [self.listbox, self.saved_items_listbox]:
            lb.search_terms = search_terms
            lb.invalidate_filter()

    def on_saved_item_deleted(self, deleted_link):
        for row in self.listbox.get_children():
            if row.feeditem.link == deleted_link:
                with row.popover.save_btn.handler_block(
                        row.popover.save_btn_handler_id
                ):
                    row.popover.save_btn.set_active(False)
                break

    def select_next_article(self, *args):
        visible_child = self.get_visible_child()
        visible_child.select_next_article()

    def select_prev_article(self, *args):
        visible_child = self.get_visible_child()
        visible_child.select_prev_article()

    def on_feeds_items_pop(self, feeditem):
        for row in self.listbox.get_children():
            if row.feeditem == feeditem:
                self.listbox.remove(row)
                break

    def on_feeds_items_append(self, feeditem):
        self.listbox.add(
            GFeedsSidebarRow(feeditem)
        )

    def on_saved_feeds_items_append(self, feeditem):
        self.saved_items_listbox.add(
            GFeedsSidebarRow(feeditem, is_saved=True)
        )

    def on_saved_feeds_items_pop(self, feeditem):
        for row in self.saved_items_listbox.get_children():
            if row.feeditem == feeditem:
                self.listbox.remove(row)
                break



from gettext import gettext as _
import threading
from feedparser import FeedParserDict
from gi.repository import Gtk, GLib, WebKit2, GObject, Gio
from gfeeds.build_reader_html import build_reader_html
from gfeeds.confManager import ConfManager
from gfeeds.download_manager import download_text
from gfeeds.revealer_loading_bar import RevealerLoadingBar


class GFeedsWebView(Gtk.Stack):
    __gsignals__ = {
        'gfeeds_webview_load_end': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        ),
        'gfeeds_webview_load_start': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        )
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_hexpand(True)
        self.set_size_request(360, 500)

        self.filler_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/webview_filler.glade'
        )
        self.webview_notif_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/webview_with_notification.glade'
        )

        self.webkitview = WebKit2.WebView()
        self.loading_bar = RevealerLoadingBar()
        self.webview_notif_builder.get_object(
            'box1'
        ).pack_start(self.loading_bar, False, False, 0)
        self.webview_notif_builder.get_object(
            'box1'
        ).pack_end(self.webkitview, True, True, 0)
        self.overlay_container = self.webview_notif_builder.get_object(
            'overlay1'
        )
        self.notif_revealer = self.webview_notif_builder.get_object(
            'revealer'
        )
        self.webview_notif_builder.get_object(
            'notif_close_btn'
        ).connect('clicked', self.hide_notif)

        self.webkitview_settings = WebKit2.Settings()
        self.apply_webview_settings()
        self.confman.connect(
            'gfeeds_webview_settings_changed',
            self.apply_webview_settings
        )

        self.webkitview.connect('load-changed', self.on_load_changed)
        self.webkitview.connect("decide-policy", self.on_decide_policy)

        self.fillerview = self.filler_builder.get_object('webview_filler_box')

        self.webkitview.set_hexpand(True)
        self.fillerview.set_hexpand(True)
        self.webkitview.set_size_request(360, 500)
        self.fillerview.set_size_request(360, 500)

        self.add_titled(self.overlay_container, 'Web View', _('Web View'))
        self.add_titled(self.fillerview, 'Filler View', _('Filler View'))
        self.set_visible_child(self.fillerview)

        self.new_page_loaded = False
        self.uri = ''
        self.feeditem = None
        self.html = None

    def apply_webview_settings(self, *args):
        self.webkitview_settings.set_enable_javascript(
            self.confman.conf['enable_js']
        )
        self.webkitview_settings.set_enable_smooth_scrolling(True)
        self.webkitview_settings.set_enable_page_cache(True)
        self.webkitview_settings.set_enable_frame_flattening(True)
        self.webkitview.set_settings(self.webkitview_settings)

    def key_zoom_in(self, *args):
        self.webkitview.set_zoom_level(
            self.webkitview.get_zoom_level()+0.15
        )

    def key_zoom_out(self, *args):
        self.webkitview.set_zoom_level(
            self.webkitview.get_zoom_level()-0.15
        )

    def key_zoom_reset(self, *args):
        self.webkitview.set_zoom_level(1.0)

    def show_notif(self, *args):
        self.notif_revealer.set_reveal_child(True)
        GLib.timeout_add_seconds(5, self.hide_notif)

    def hide_notif(self, *args):
        self.notif_revealer.set_reveal_child(False)

    def set_enable_rss_content(self, state=True, feeditem=None):
        if feeditem:
            self.feeditem = feeditem
        if state:
            self._load_rss_content(self.feeditem)
        else:
            self.new_page_loaded = True
            self.load_feeditem(
                self.feeditem,
                False
            )

    def _load_rss_content(self, feeditem):
        if feeditem.is_saved:
            self.set_enable_reader_mode(True, False)
            return
        self.set_visible_child(self.overlay_container)
        self.feeditem = feeditem
        self.uri = feeditem.link
        content = feeditem.fp_item.get(
            'content', feeditem.fp_item.get('summary', None)
        )
        if not content:
            content = '<h1><i>'+_(
                'RSS content or summary not available for this article'
                )+'</i></h1>'
        elif not isinstance(content, str):
            if isinstance(content, list):
                content = content[0]
            if isinstance(content, (dict, FeedParserDict)):
                if 'value' in content.keys():
                    content = content['value']
        self.html = '<!-- GFEEDS RSS CONTENT --><article>{0}</article>'.format(
            content if '</' in content else content.replace('\n', '<br>')
        )
        self.set_enable_reader_mode(True, True)

    def _load_reader_async(self, callback=None, *args):
        self.html = download_text(self.uri)
        GLib.idle_add(
            self.set_enable_reader_mode, True
        )
        if callback:
            GLib.idle_add(callback)

    def load_feeditem(self, feeditem, trigger_on_load_start=True,
                      *args, **kwargs):
        self.webkitview.stop_loading()
        uri = feeditem.link
        if feeditem.is_saved:
            uri = (
                'file://' +
                self.confman.saved_cache_path + '/' +
                feeditem.fp_item['linkhash']
            )
        self.feeditem = feeditem
        self.uri = uri
        self.set_visible_child(self.overlay_container)
        if self.confman.conf['default_view'] == 'reader':
            t = threading.Thread(
                group=None,
                target=self._load_reader_async,
                name=None,
                # args = (uri,)
            )
            if trigger_on_load_start:
                self.on_load_start()
            t.start()
        elif self.confman.conf['default_view'] == 'rsscont':
            self.on_load_start()
            self.set_enable_rss_content(True, feeditem)
        else:
            self.webkitview.load_uri(uri)
            if trigger_on_load_start:
                self.on_load_start()

    def open_externally(self, *args):
        if self.uri:
            Gio.AppInfo.launch_default_for_uri(
                self.uri
            )

    def on_load_start(self, *args):
        self.new_page_loaded = True
        self.emit('gfeeds_webview_load_start', '')

    def on_load_changed(self, webview, event):
        if event != WebKit2.LoadEvent.FINISHED:
            self.loading_bar.set_reveal_child(True)
        if event == WebKit2.LoadEvent.STARTED:
            self.loading_bar.set_fraction(.25)
        elif event == WebKit2.LoadEvent.REDIRECTED:
            self.loading_bar.set_fraction(.50)
        elif event == WebKit2.LoadEvent.COMMITTED:
            self.emit('gfeeds_webview_load_end', '')
            self.loading_bar.set_fraction(.75)
        elif self.new_page_loaded and event == WebKit2.LoadEvent.FINISHED:
            self.loading_bar.set_fraction(1.0)
            # waits 3 seconds async then hides the loading bar
            GLib.timeout_add_seconds(
                3,
                self.loading_bar.set_reveal_child, False
            )
            self.new_page_loaded = False
            resource = webview.get_main_resource()
            resource.get_data(None, self._get_data_cb, None)

    def _set_enable_reader_mode_async_callback(self):
        self.webkitview.load_html(build_reader_html(
            self.html,
            self.confman.conf['dark_reader'],
            self.feeditem.fp_item
        ), self.uri)

    def set_enable_reader_mode(self, state=True, is_rss_content=False):
        if state:
            if (
                    not self.html or (
                        not is_rss_content and
                        (
                            self.html[:36] ==
                            '<!-- GFEEDS RSS CONTENT --><article>'
                        )
                    )
            ):
                t = threading.Thread(
                    group=None,
                    target=self._load_reader_async,
                    name=None,
                    args=(
                        self._set_enable_reader_mode_async_callback,
                    )
                )
                t.start()
            else:
                self._set_enable_reader_mode_async_callback()
        else:
            self.webkitview.load_uri(self.uri)

    def _get_data_cb(self, resource, result, user_data=None):
        self.html = resource.get_data_finish(result)

    def on_decide_policy(self, webView, decision, decisionType):
        if not self.confman.conf['open_links_externally']:
            return False
        if (
                decisionType in
                (
                    WebKit2.PolicyDecisionType.NAVIGATION_ACTION,
                    WebKit2.PolicyDecisionType.NEW_WINDOW_ACTION
                ) and
                decision.get_navigation_action().get_mouse_button() != 0
        ):
            decision.ignore()
            uri = decision.get_navigation_action().get_request().get_uri()
            Gio.AppInfo.launch_default_for_uri(uri)
            return True
        return False



from pathlib import Path
from os.path import isfile, isdir
from os import makedirs
from os import environ as Env
import json
from datetime import timedelta
from gi.repository import Gtk, GObject, Gio
from gfeeds.singleton import Singleton
from gfeeds.signaler_list import SignalerList


class ConfManagerSignaler(GObject.Object):

    __gsignals__ = {
        'gfeeds_new_first_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_repopulation_required': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_webview_settings_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_show_read_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_full_article_title_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        # Signals down here don't have to do with the config
        'gfeeds_filter_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (GObject.TYPE_PYOBJECT,)
        ),
        'gfeeds_full_feed_name_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_tags_append': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'gfeeds_tags_pop': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
    }


class ConfManager(metaclass=Singleton):

    _background_color = None

    BASE_SCHEMA = {
        'feeds': {},
        'dark_reader': False,
        'new_first': True,
        'windowsize': {
            'width': 350,
            'height': 650
        },
        'max_article_age_days': 30,
        'enable_js': False,
        'max_refresh_threads': 2,
        'saved_items': {},
        'read_items': [],
        'show_read_items': True,
        'full_article_title': True,
        # valid values: 'webview', 'reader', 'rsscont'
        'default_view': 'webview',
        'open_links_externally': True,
        'full_feed_name': False,
        'refresh_on_startup': False,
        'tags': [],
        'open_youtube_externally': False,
        'media_player': 'mpv'
    }

    def __init__(self):
        self.window = None
        self.signaler = ConfManagerSignaler()
        self.emit = self.signaler.emit
        self.connect = self.signaler.connect

        # check if inside flatpak sandbox
        self.is_flatpak = (
            'XDG_RUNTIME_DIR' in Env.keys() and
            isfile(f'{Env["XDG_RUNTIME_DIR"]}/flatpak-info')
        )

        if self.is_flatpak:
            self.conf_dir = Env.get("XDG_CONFIG_HOME")
            self.cache_path = Path(
                f'{Env.get("XDG_CACHE_HOME")}/org.gabmus.gfeeds'
            )
        else:
            self.conf_dir = f'{Env.get("HOME")}/.config'
            self.cache_path = Path(
                f'{Env.get("HOME")}/.cache/org.gabmus.gfeeds'
            )
        self.path = Path(
            f'{self.conf_dir}/org.gabmus.gfeeds.json'
        )
        self.thumbs_cache_path = f'{self.cache_path}/thumbnails/'
        old_saved_cache_path = f'{self.cache_path}/saved_articles'
        self.saved_cache_path = \
            f'{self.conf_dir}/org.gabmus.gfeeds.saved_articles'

        self.conf = None
        if self.path.is_file():
            try:
                with open(self.path) as fd:
                    self.conf = json.loads(fd.read())
                # verify that the file has all of the schema keys
                for k in ConfManager.BASE_SCHEMA:
                    if k not in self.conf.keys():
                        if isinstance(
                                ConfManager.BASE_SCHEMA[k], (list, dict)
                        ):
                            self.conf[k] = ConfManager.BASE_SCHEMA[k].copy()
                        else:
                            self.conf[k] = ConfManager.BASE_SCHEMA[k]
                if isinstance(self.conf['feeds'], list):
                    n_feeds = {}
                    for o_feed in self.conf['feeds']:
                        n_feeds[o_feed] = {}
                    self.conf['feeds'] = n_feeds
                    self.save_conf()
            except Exception:
                self.conf = ConfManager.BASE_SCHEMA.copy()
                self.save_conf()
        else:
            self.conf = ConfManager.BASE_SCHEMA.copy()
            self.save_conf()

        for p in [
                self.cache_path,
                self.thumbs_cache_path,
                self.saved_cache_path
        ]:
            if not isdir(str(p)):
                makedirs(str(p))
        # TODO: remove down the road, old saved articles migration
        if isdir(old_saved_cache_path):
            from os import listdir, rmdir
            from shutil import move
            for f in listdir(old_saved_cache_path):
                move(
                    f'{old_saved_cache_path}/{f}',
                    f'{self.saved_cache_path}/{f}'
                )
            rmdir(old_saved_cache_path)

        self.read_feeds_items = SignalerList(self.conf['read_items'])
        self.read_feeds_items.connect(
            'append', self.dump_read_items_to_conf
        )
        self.read_feeds_items.connect(
            'pop', self.dump_read_items_to_conf
        )

        bl_gsettings = Gio.Settings.new('org.gnome.desktop.wm.preferences')
        bl = bl_gsettings.get_value('button-layout').get_string()
        self.wm_decoration_on_left = (
            'close:' in bl or
            'maximize:' in bl or
            'minimize:' in bl
        )

        # font_gsettings = Gio.Settings.new('org.gnome.destkop.interface')
        # self.sans_font = font_gsettings.get_value(
        #     'font-name'
        # ).get_string()
        # self.serif_font = font_gsettings.get_value(
        #     'document-font-name'
        # ).get_string()
        # self.mono_font = font_gsettings.get_value(
        #     'monospace-font-name'
        # ).get_string()

    @property
    def max_article_age(self) -> timedelta:
        return timedelta(days=self.conf['max_article_age_days'])

    def add_tag(self, tag: str, target_feeds=[]):
        lowercase_tags = [t.lower() for t in self.conf['tags']]
        if tag.lower() not in lowercase_tags:
            self.conf['tags'].append(tag)
            self.emit('gfeeds_tags_append', tag)
        for feed in target_feeds:
            if 'tags' not in self.conf['feeds'][feed].keys():
                self.conf['feeds'][feed]['tags'] = []
            if tag not in self.conf['feeds'][feed]['tags']:
                self.conf['feeds'][feed]['tags'].append(tag)
        self.save_conf()

    def delete_tag(self, tag: str):
        while tag in self.conf['tags']:
            self.conf['tags'].remove(tag)
        self.emit('gfeeds_tags_pop', tag)
        print(self.conf['tags'])
        self.remove_tag(tag, self.conf['feeds'].keys())
        # self.save_conf()  # done by remove_tags

    def remove_tag(self, tag: str, target_feeds: list):
        for feed in target_feeds:
            if 'tags' not in self.conf['feeds'][feed].keys():
                continue
            if tag in self.conf['feeds'][feed]['tags']:
                self.conf['feeds'][feed]['tags'].remove(tag)
        self.save_conf()

    def dump_read_items_to_conf(self, *args):
        self.conf['read_items'] = self.read_feeds_items.get_list()

    def save_conf(self, *args):
        if self.path.is_file():
            with open(self.path, 'r') as fd:
                if json.loads(fd.read()) == self.conf:
                    return
        with open(self.path, 'w') as fd:
            fd.write(json.dumps(self.conf))

    def get_background_color(self) -> str:
        if ConfManager._background_color is not None:
            return ConfManager._background_color
        if not self.window:
            return "000000"
        gc = self.window.get_style_context(
                ).get_background_color(Gtk.StateFlags.NORMAL)
        color = ''
        for channel in (gc.red, gc.green, gc.blue):
            color += '%02x' % int(channel*255)
        ConfManager._background_color = color
        return color



# __main__.py
#
# Copyright (C) 2019 GabMus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import argparse
from gettext import gettext as _
from os.path import isfile
from gi.repository import Gtk, Gdk, Gio, GLib, Handy
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.app_window import GFeedsAppWindow
from gfeeds.settings_window import show_settings_window
from gfeeds.opml_manager import (
    feeds_list_to_opml,
    add_feeds_from_opml
)
from gfeeds.opml_file_chooser import (
    GFeedsOpmlFileChooserDialog,
    GFeedsOpmlSavePathChooserDialog
)
from gfeeds.manage_feeds_window import GFeedsManageFeedsWindow
from gfeeds.confirm_add_dialog import GFeedsConfirmAddDialog
from gfeeds.shortcuts_window import show_shortcuts_window
from gfeeds.rss_link_from_file import get_feed_link_from_file


class GFeedsApplication(Gtk.Application):
    def __init__(self, **kwargs):
        super().__init__(
            application_id='org.gabmus.gfeeds',
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            **kwargs
        )
        GLib.set_application_name('Feeds')
        GLib.set_prgname('org.gabmus.gfeeds')
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.window = GFeedsAppWindow()
        self.confman.window = self.window
        self.window.connect('destroy', self.on_destroy_window)

    def do_startup(self):
        Gtk.Application.do_startup(self)
        Handy.init()
        self.feedman.refresh(
            get_cached=not self.confman.conf['refresh_on_startup'],
            is_startup=True
        )

        stateful_actions = [
            {
                'name': 'show_read_items',
                'func': self.show_read_items,
                'type': 'bool',
                'accel': '<Control>h',
                'confman_key': 'show_read_items'
            },
            {
                'name': 'view_mode_change',
                'func': self.view_mode_change,
                'type': 'radio',
                'confman_key': 'default_view'
            }
        ]

        actions = [
            {
                'name': 'set_all_read',
                'func': self.set_all_read
            },
            {
                'name': 'set_all_unread',
                'func': self.set_all_unread
            },
            {
                'name': 'manage_feeds',
                'func': self.manage_feeds
            },
            {
                'name': 'import_opml',
                'func': self.import_opml
            },
            {
                'name': 'export_opml',
                'func': self.export_opml
            },
            {
                'name': 'settings',
                'func': lambda *args: show_settings_window(self.window),
                'accel': '<Primary>comma'
            },
            {
                'name': 'shortcuts',
                'func': lambda *args: show_shortcuts_window(self.window),
                'accel': '<Primary>question'
            },
            {
                'name': 'about',
                'func': self.show_about_dialog
            },
            {
                'name': 'quit',
                'func': self.on_destroy_window,
                'accel': '<Primary>q'
            }
        ]

        for sa in stateful_actions:
            c_action = None
            if sa['type'] == 'bool':
                c_action = Gio.SimpleAction.new_stateful(
                    sa['name'],
                    None,
                    GLib.Variant.new_boolean(
                        self.confman.conf[sa['confman_key']]
                    )
                )
            elif sa['type'] == 'radio':
                c_action = Gio.SimpleAction.new_stateful(
                    sa['name'],
                    GLib.VariantType.new('s'),
                    GLib.Variant('s', self.confman.conf[sa['confman_key']])
                )
            else:
                raise ValueError(
                    f'Stateful Action: unsupported type `{sa["type"]}`'
                )
            c_action.connect('activate', sa['func'])
            self.add_action(c_action)
            if 'accel' in sa.keys():
                self.set_accels_for_action(
                    f'app.{sa["name"]}',
                    [sa['accel']]
                )

        for a in actions:
            c_action = Gio.SimpleAction.new(a['name'], None)
            c_action.connect('activate', a['func'])
            self.add_action(c_action)
            if 'accel' in a.keys():
                self.set_accels_for_action(
                    f'app.{a["name"]}',
                    [a['accel']]
                )

    def view_mode_change(
            self,
            action: Gio.SimpleAction,
            target: GLib.Variant,
            *args
    ):
        action.change_state(target)
        target_s = str(target).strip("'")
        if target_s not in ['webview', 'reader', 'rsscont']:
            target_s = 'webview'
        self.window.headerbar.on_view_mode_change(target_s)
        self.confman.conf['default_view'] = target_s
        self.confman.save_conf()

    def show_read_items(self, action: Gio.SimpleAction, *args):
        action.change_state(
            GLib.Variant.new_boolean(not action.get_state().get_boolean())
        )
        self.confman.conf['show_read_items'] = action.get_state().get_boolean()
        self.confman.emit('gfeeds_show_read_changed', '')

    def set_all_read(self, *args):
        for row in self.window.sidebar.listbox.get_children():
            # check if row is visible in listbox
            if self.window.sidebar.listbox.gfeeds_sidebar_filter_func(
                    row, None, None
            ):
                row.popover.set_read(True)

    def set_all_unread(self, *args):
        for row in self.window.sidebar.listbox.get_children():
            row.popover.set_read(False)

    def manage_feeds(self, *args):
        mf_win = GFeedsManageFeedsWindow(
            self.window
        )
        mf_win.present()

    def import_opml(self, *args):
        dialog = GFeedsOpmlFileChooserDialog(self.window)
        res = dialog.run()
        # dialog.close()
        if res == Gtk.ResponseType.ACCEPT:
            add_feeds_from_opml(dialog.get_filename())

    def export_opml(self, *args):
        dialog = GFeedsOpmlSavePathChooserDialog(self.window)
        res = dialog.run()
        # dialog.close()
        if res == Gtk.ResponseType.ACCEPT:
            save_path = dialog.get_filename()
            if save_path[-5:].lower() != '.opml':
                save_path += '.opml'
            opml_out = feeds_list_to_opml(self.feedman.feeds)
            with open(save_path, 'w') as fd:
                fd.write(opml_out)
                fd.close()

    def show_about_dialog(self, *args):
        about_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/aboutdialog.glade'
        )
        dialog = about_builder.get_object('aboutdialog')
        dialog.set_modal(True)
        dialog.set_transient_for(self.window)
        dialog.present()

    def on_destroy_window(self, *args):
        self.window.on_destroy()
        self.quit()

    def do_activate(self):
        self.add_window(self.window)
        stylecontext = Gtk.StyleContext()
        provider = Gtk.CssProvider()
        provider.load_from_resource(
            '/org/gabmus/gfeeds/ui/gtk_style.css'
        )
        stylecontext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
        self.window.present()
        self.window.show_all()
        # self.feedman.refresh(get_cached=True)
        if self.args:
            if self.args.argurl:
                if self.args.argurl[:8].lower() == 'file:///':
                    abspath = self.args.argurl[7:]
                    if isfile(abspath):
                        if abspath[-5:].lower() == '.opml':
                            dialog = GFeedsConfirmAddDialog(
                                self.window, abspath
                            )
                            res = dialog.run()
                            dialog.close()
                            if res == Gtk.ResponseType.YES:
                                add_feeds_from_opml(abspath)
                        else:
                            # why no check for extension here?
                            # some websites have feeds without extension
                            # dumb but that's what it is
                            self.args.argurl = get_feed_link_from_file(
                                abspath
                            ) or ''
                if (
                        self.args.argurl[:7].lower() == 'http://' or
                        self.args.argurl[:8].lower() == 'https://'
                ):
                    dialog = GFeedsConfirmAddDialog(
                        self.window,
                        self.args.argurl,
                        http=True
                    )
                    res = dialog.run()
                    dialog.close()
                    if res == Gtk.ResponseType.YES:
                        self.feedman.add_feed(self.args.argurl)
                else:
                    print('This file is not supported')

    def do_command_line(self, args: list):
        """
        GTK.Application command line handler
        called if Gio.ApplicationFlags.HANDLES_COMMAND_LINE is set.
        must call the self.do_activate() to get the application up and running.
        """
        # call the default commandline handler
        Gtk.Application.do_command_line(self, args)
        # make a command line parser
        parser = argparse.ArgumentParser()
        parser.add_argument(
            'argurl',
            metavar=_('url'),
            type=str,
            nargs='?',
            help=_('opml file local url or rss remote url to import')
        )
        # parse the command line stored in args,
        # but skip the first element (the filename)
        self.args = parser.parse_args(args.get_arguments()[1:])
        # call the main program do_activate() to start up the app
        self.do_activate()
        return 0


def main():

    application = GFeedsApplication()

    try:
        ret = application.run(sys.argv)
    except SystemExit as e:
        ret = e.code

    sys.exit(ret)


if __name__ == '__main__':
    main()



from gettext import gettext as _
from xml.sax.saxutils import escape
from gi.repository import Gtk, Pango
from gfeeds.feeds_manager import FeedsManager
from gfeeds.confManager import ConfManager
from gfeeds.scrolled_message_dialog import ScrolledMessageDialog


class GFeedsInfoBar(Gtk.InfoBar):
    def __init__(self, text, icon_name=None,
                 message_type=Gtk.MessageType.INFO, **kwargs):
        super().__init__(**kwargs)
        self.set_message_type(message_type)
        self.text = text
        self.container_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.label = Gtk.Label(label=self.text)
        self.label.set_line_wrap(True)
        self.label.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.label.set_xalign(0.0)
        if icon_name:
            self.icon_name = icon_name
            self.icon = Gtk.Image.new_from_icon_name(
                self.icon_name,
                Gtk.IconSize.BUTTON
            )
            self.container_box.pack_start(self.icon, False, False, 6)
        self.container_box.pack_start(self.label, False, False, 6)
        self.container_box.set_halign(Gtk.Align.CENTER)
        self.container_box.set_margin_top(6)
        self.container_box.set_margin_bottom(6)
        self.get_content_area().set_center_widget(self.container_box)
        self.set_hexpand(False)
        # self.set_size_request(360, -1)


class GFeedsErrorsBar(GFeedsInfoBar):
    def __init__(self, parent_win, **kwargs):
        super().__init__(
            text=_('There are some errors'),
            icon_name='computer-fail-symbolic',
            message_type=Gtk.MessageType.INFO,
            **kwargs
        )
        self.confman = ConfManager()
        self.parent_win = parent_win
        self.errors = []
        self.problematic_feeds = []
        self.show_button = Gtk.Button(label=_('Show'))
        self.ignore_button = Gtk.Button(label=_('Ignore'))
        self.show_button.connect('clicked', self.show_errors)
        self.ignore_button.connect(
            'clicked',
            lambda *args: self.set_revealed(False)
        )
        self.container_box.pack_end(self.ignore_button, False, False, 6)
        self.container_box.pack_end(self.show_button, False, False, 6)
        self.set_revealed(False)

    def engage(self, errors: list, problematic_feeds: list):
        self.errors = errors
        self.problematic_feeds = problematic_feeds
        if len(errors) != 0:
            self.set_revealed(True)

    def show_errors(self, *args):
        dialog = ScrolledMessageDialog(
            self.parent_win,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.YES_NO,
            _(
                'There were problems with some feeds. '
                'Do you want to remove them?'
            )
        )
        dialog.format_secondary_markup(
            escape('\n'.join(self.errors))
        )
        if (dialog.run() == Gtk.ResponseType.YES):
            for pf in self.problematic_feeds:
                if pf in self.confman.conf['feeds'].keys():
                    self.confman.conf['feeds'].pop(pf)
            self.set_revealed(False)
        else:
            self.set_revealed(True)
        dialog.close()


class GFeedsConnectionBar(GFeedsInfoBar):
    def __init__(self, **kwargs):
        super().__init__(
            text=_('You are offline'),
            icon_name='network-offline-symbolic',
            message_type=Gtk.MessageType.INFO,
            **kwargs
        )
        self.container_box.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.feedman = FeedsManager()
        self.feedman.connect(
            'feedmanager_online_changed',
            lambda caller, value: self.set_revealed(not value)
        )
        self.set_revealed(False)



import threading
from gettext import gettext as _
from gi.repository import GLib, GObject
from gfeeds.singleton import Singleton
from gfeeds.confManager import ConfManager
from gfeeds.rss_parser import Feed, FeedItem
from gfeeds.download_manager import (
    download_feed,
    extract_feed_url_from_html
)
from gfeeds.signaler_list import SignalerList
from gfeeds.test_connection import is_online
from gfeeds.thread_pool import ThreadPool


class FeedsManagerSignaler(GObject.Object):
    __gsignals__ = {
        'feedmanager_refresh_start': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        ),
        'feedmanager_refresh_end': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        ),
        'feedmanager_online_changed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (bool,)
        ),
    }


class FeedsManager(metaclass=Singleton):
    def __init__(self):
        self.confman = ConfManager()
        self.confman.connect(
            'gfeeds_repopulation_required',
            self.refresh
        )
        self.signaler = FeedsManagerSignaler()
        self.emit = self.signaler.emit
        self.connect = self.signaler.connect

        self.feeds = SignalerList()
        self.feeds_items = SignalerList()
        self.saved_feeds_items = SignalerList()

        self.errors = []
        self.problematic_feeds = []
        self.connect(
            'feedmanager_refresh_end',
            lambda *args: self.confman.save_conf()
        )

    def populate_saved_feeds_items(self):
        self.saved_feeds_items.empty()
        for si in self.confman.conf['saved_items'].values():
            self.saved_feeds_items.append(
                FeedItem.new_from_dict(si)
            )

    def _add_feed_async_worker(
            self,
            uri: str,
            refresh: bool = False,
            get_cached: bool = False
    ):
        if not refresh:
            if 'http://' not in uri and 'https://' not in uri:
                uri = 'http://' + uri
            if uri in self.confman.conf['feeds'].keys():
                print(_('Feed {0} exists already, skipping').format(uri))
                return
            self.confman.conf['feeds'][uri] = {}
        download_res = download_feed(uri, get_cached=get_cached)
        if get_cached and download_res[0] == 'not_cached':
            return
        n_feed = Feed(download_res)
        if n_feed.is_null:
            feed_uri_from_html = extract_feed_url_from_html(uri)
            if feed_uri_from_html is not None:
                if uri in self.confman.conf['feeds'].keys():
                    self.confman.conf['feeds'].pop(uri)
                self._add_feed_async_worker(feed_uri_from_html, refresh)
                return
            self.errors.append(n_feed.error)
            self.problematic_feeds.append(uri)
        else:
            GLib.idle_add(
                self.feeds.append, n_feed, priority=GLib.PRIORITY_LOW
            )
            GLib.idle_add(
                self.feeds_items.extend,
                n_feed.items,
                priority=GLib.PRIORITY_LOW
            )
        if not refresh:
            GLib.idle_add(
                self.emit, 'feedmanager_refresh_end', ''
            )

    def refresh(
            self,
            *args,
            get_cached: bool = False,
            is_startup: bool = False
    ):
        self.emit(
            'feedmanager_refresh_start',
            'startup' if is_startup else ''
        )
        self.errors = []
        self.problematic_feeds = []
        if is_online():
            self.emit('feedmanager_online_changed', True)
        else:
            self.emit('feedmanager_online_changed', False)
            get_cached = True
            # self.emit('feedmanager_refresh_end', '')
            # return
        self.populate_saved_feeds_items()
        self.feeds.empty()
        self.feeds_items.empty()
        tp = ThreadPool(
            self.confman.conf['max_refresh_threads'],
            self._add_feed_async_worker,
            [
                (f_link, True, get_cached)
                for f_link in self.confman.conf['feeds'].keys()
            ],
            self.emit,
            ('feedmanager_refresh_end', '')
        )
        tp.start()

    def add_feed(self, uri: str, is_new: bool = False) -> bool:
        if is_new and uri in self.confman.conf['feeds'].keys():
            return False
        self.emit('feedmanager_refresh_start', '')
        self.errors = []
        t = threading.Thread(
            group=None,
            target=self._add_feed_async_worker,
            name=None,
            args=(uri,)
        )
        t.start()
        return True

    def delete_feeds(self, targets, *args):
        if type(targets) != list:
            if type(targets) == Feed:
                targets = [targets]
            else:
                raise TypeError('delete_feed: targets must be list or Feed')
        for f in targets:
            for fi in f.items:
                self.feeds_items.remove(fi)
            self.feeds.remove(f)
            self.confman.conf['feeds'].pop(
                f.rss_link
            )
        self.confman.save_conf()



from gi.repository import Gtk


class RevealerLoadingBar(Gtk.Revealer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.progress_bar = Gtk.ProgressBar()
        self.add(self.progress_bar)
        self.set_fraction = self.progress_bar.set_fraction
        self.set_reveal_child(True)



from gi.repository import Gtk


class GFeedsViewModeMenu(Gtk.PopoverMenu):
    def __init__(self, relative_to, **kwargs):
        super().__init__(**kwargs)
        self.set_modal(True)
        self.set_relative_to(relative_to)
        relative_to.set_popover(self)
        relative_to.set_sensitive(False)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/extra_popover_menu.glade'
        )
        self.add(self.builder.get_object(
            'menu_box'
        ))



from typing import Tuple
from gettext import gettext as _
import readability
import pygments
import pygments.lexers
from lxml.html import (
    fromstring as html_fromstring,
    tostring as html_tostring,
    HtmlElement
)
from pygments.formatters import HtmlFormatter
from gfeeds.reader_mode_style import CSS, DARK_MODE_CSS


# Thanks to Eloi Rivard (azmeuk) for the contribution on the media block
def _build_media_text(title: str, content: str) -> str:
    return '''
        <p>
            <strong>{0}:</strong>
            {1}
        </p>'''.format(title, content.replace("\n", "<br />"))


def _build_media_link(title: str, content: str, link: str):
    return _build_media_text(
        title, f'<a href="{link}">{content}</a>'
    )


# thumbnails aren't supposed to be links, images can be?
# funnily enough, using `#` as a link opens the main content url
# that's because the webkitview sets the base url to the feed item link
def _build_media_img(title, imgurl, link='#') -> str:
    return _build_media_link(
        title, f'<br /><img src="{imgurl}" />', link
    )


def build_syntax_highlight(root: HtmlElement) -> Tuple[str, HtmlElement]:
    syntax_highlight_css = ''
    code_nodes = root.xpath(
        '//pre/code'
    )
    lexer = None
    for code_node in code_nodes:
        classes = code_node.attrib.get('class', '').split(' ')
        for klass in classes:
            try:
                lexer = pygments.lexers.get_lexer_by_name(
                    klass.replace('language-', ''),
                    stripall=True,
                )
                break
            except pygments.util.ClassNotFound:
                pass

        if not code_node.text:
            continue

        if not lexer:
            try:
                lexer = pygments.lexers.guess_lexer(code_node.text)
            except pygments.util.ClassNotFound:
                continue

        formatter = HtmlFormatter(style='solarized-dark')

        if not syntax_highlight_css:
            syntax_highlight_css = formatter.get_style_defs()

        newtext = pygments.highlight(code_node.text, lexer, formatter)
        newhtml = html_fromstring(newtext)

        pre_node = code_node.getparent()
        pre_node.getparent().replace(pre_node, newhtml)

    return syntax_highlight_css, root


def build_syntax_highlight_from_raw_html(raw_html) -> Tuple[str, HtmlElement]:
    return build_syntax_highlight(
        html_fromstring(
            raw_html if type(raw_html) == str else raw_html.decode()
        )
    )


def build_reader_html(og_html, dark_mode: bool = False, fp_item=None) -> str:
    def build_media_block():
        if not fp_item:
            return ''
        media_s = '<hr />'
        for el in fp_item:
            if el[:6].lower() == 'media_':
                try:
                    if el.lower() == 'media_thumbnail':
                        media_s += _build_media_img(
                            _('Thumbnail'),
                            fp_item["media_thumbnail"][0]["url"]
                        )
                    else:
                        for subel in fp_item[el]:
                            media_s += _build_media_text(
                                subel.capitalize(),
                                fp_item[el][subel]
                            )
                except Exception:
                    continue
        return media_s if media_s != '<hr />' else ''

    doc = readability.Document(og_html)
    content = doc.summary(True)
    syntax_highlight_css, root = build_syntax_highlight_from_raw_html(content)
    content = html_tostring(
        root, encoding='utf-8'
    ).decode().replace(
        '</p><a ', '<a '
    ).replace(
        '</a><p>', '</a>'
    ).replace(
        '</p><em>', '<em>'
    ).replace(
        '</em><p>', '</em>'
    )
    content += build_media_block()
    return f'''<html>
        <head>
            <meta charset="UTF-8">
            <style>
                {CSS}
                {DARK_MODE_CSS if dark_mode else ""}
                {syntax_highlight_css}
            </style>
            <title>{doc.short_title()}</title>
        </head>
        <body>
            <article>
                <h1>{doc.short_title() or fp_item["title"]}</h1>
                {content}
            </article>
        </body>
    </html>'''



from gettext import gettext as _
from urllib.parse import urlparse
from lxml.html import html5parser
from PIL import Image
from gfeeds.download_manager import download_raw, download_text


def get_favicon(link, favicon_path):
    # req = requests.get(link)
    # if not 200 <= req.status_code <= 299:
    #     print(f'get_favicon: failed with code {req.status_code}')
    #     return None
    html = download_text(link)
    root = html5parser.fromstring(
        html if isinstance(html, str) else html.decode()
    )
    favicon_els = root.xpath(
        '//x:link',
        namespaces={'x': 'http://www.w3.org/1999/xhtml'}
    )
    candidate = {
        'path': '',
        'is_absolute': False,
        'size': -1
    }
    for e in favicon_els:
        if candidate['size'] >= 100:
            break
        if 'rel' in e.attrib.keys():
            size = 0
            can_save = False
            if e.attrib['rel'] == 'apple-touch-icon':
                size = 1
                can_save = True
                if 'sizes' in e.attrib.keys():
                    size = int(e.attrib['sizes'].split('x')[0])
            elif (
                    e.attrib['rel'] in ['icon', 'shortcut icon'] and
                    candidate['size'] == -1
            ):
                size = 0
                can_save = True
            if can_save:
                candidate['path'] = e.attrib['href']
                candidate['is_absolute'] = (
                    'http://' in e.attrib['href'] or
                    'https://' in e.attrib['href']
                )
                candidate['size'] = size
    p = candidate['path']
    needs_convert = False
    if not p:
        return
    if p[-4:].lower() == '.ico':
        needs_convert = True
    if not candidate['is_absolute']:
        if p[0:2] == '//':
            p = p[2:]
        elif p[0] == '/':
            p = p[1:]
        up = urlparse(link)
        url = f'{up.scheme or "http"}://{up.hostname}/{p}'
        try:
            download_raw(url, favicon_path)
        except Exception:
            try:
                url = f'{up.scheme or "http"}://{p}'
                download_raw(url, favicon_path)
            except Exception:
                print(_('Error downloading favicon for `{0}`').format(link))
    else:
        download_raw(p, favicon_path)
    if needs_convert:
        toconv = Image.open(favicon_path)
        toconv.save(favicon_path)
        toconv.close()



from gettext import gettext as _
from gi.repository import Gtk, Gdk, Handy, GObject
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.spinner_button import RefreshSpinnerButton
from gfeeds.feeds_view import FeedsViewPopover
from gfeeds.view_mode_menu import GFeedsViewModeMenu


class AddFeedPopover(Gtk.Popover):
    def __init__(self, relative_to, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()
        self.feedman = FeedsManager()

        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/add_feed_box.glade'
        )
        self.set_modal(True)
        self.set_relative_to(relative_to)
        relative_to.connect('clicked', self.on_relative_to_clicked)
        self.container_box = self.builder.get_object('container_box')
        self.confirm_btn = self.builder.get_object('confirm_btn')
        self.confirm_btn.connect(
            'clicked',
            self.on_confirm_btn_clicked
        )
        self.url_entry = self.builder.get_object('url_entry')
        self.url_entry.connect('activate', self.on_url_entry_activate)
        self.already_subscribed_revealer = self.builder.get_object(
            'already_subscribed_revealer'
        )
        # about this lambda: low impact, happens rarely
        self.url_entry.connect(
            'changed',
            lambda *args:
            self.already_subscribed_revealer.set_reveal_child(False)
        )
        self.add(self.container_box)

    def on_url_entry_activate(self, *args):
        if self.confirm_btn.get_sensitive():
            self.on_confirm_btn_clicked(self.confirm_btn)

    def on_relative_to_clicked(self, *args):
        self.popup()

    def on_confirm_btn_clicked(self, btn):
        res = self.feedman.add_feed(self.url_entry.get_text(), True)
        if res:
            self.popdown()
            self.already_subscribed_revealer.set_reveal_child(False)
        else:
            self.already_subscribed_revealer.set_reveal_child(True)


class GFeedHeaderbar(Handy.WindowHandle):
    __gsignals__ = {
        'gfeeds_headerbar_squeeze': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (bool,)
        )
    }

    def __init__(
            self,
            size_group_left,
            size_group_right,
            back_btn_func,
            webview,
            searchbar,
            **kwargs):
        super().__init__(**kwargs)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/headerbar.glade'
        )
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.back_btn_func = back_btn_func
        self.webview = webview
        self.searchbar = searchbar
        self.webview.connect('gfeeds_webview_load_start', self.on_load_start)
        self.webview.connect('gfeeds_webview_load_end', self.on_load_end)
        self.headergroup = Handy.HeaderGroup()
        leaflet_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/gfeeds_leaflet.glade'
        )
        self.leaflet = leaflet_builder.get_object('leaflet')
        self.leaflet.set_can_swipe_back(False)
        self.left_headerbar = self.builder.get_object(
            'left_headerbar'
        )
        self.right_headerbar = self.builder.get_object(
            'right_headerbar'
        )
        self.right_headerbar.set_hexpand(True)
        size_group_left.add_widget(self.left_headerbar)
        size_group_right.add_widget(self.right_headerbar)
        self.headergroup.add_header_bar(self.left_headerbar)
        self.headergroup.add_header_bar(self.right_headerbar)
        separator = Gtk.Separator()
        separator.get_style_context().add_class('sidebar')

        self.leaflet.add(self.left_headerbar)
        self.leaflet.add(separator)
        self.leaflet.add(self.right_headerbar)
        self.leaflet.child_set_property(separator, 'allow-visible', False)
        self.add(self.leaflet)
        self.set_headerbar_controls()
        # self.headergroup.set_focus(self.left_headerbar)

        self.back_button = self.builder.get_object('back_btn')
        self.back_button.connect('clicked', self.on_back_button_clicked)
        self.view_mode_menu_btn = self.builder.get_object(
            'view_mode_menu_btn'
        )
        self.view_mode_menu_btn_icon = self.builder.get_object(
            'view_mode_menu_btn_icon'
        )
        self.set_view_mode_icon(self.confman.conf['default_view'])
        self.view_mode_menu = GFeedsViewModeMenu(self.view_mode_menu_btn)
        self.open_externally_btn = self.builder.get_object(
            'open_externally_btn'
        )
        self.open_externally_btn.connect(
            'clicked', self.webview.open_externally
        )
        self.share_btn = self.builder.get_object('share_btn')
        self.share_btn.connect('clicked', self.copy_article_uri)
        self.menu_btn = self.builder.get_object(
            'menu_btn'
        )
        self.menu_popover = Gtk.PopoverMenu()
        self.menu_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/menu.xml'
        )
        self.menu = self.menu_builder.get_object('generalMenu')
        self.menu_popover.bind_model(self.menu)
        self.menu_btn.set_popover(self.menu_popover)

        self.search_btn = self.builder.get_object('search_btn')
        self.search_btn.connect('toggled', self.on_search_btn_toggled)
        self.filter_btn = self.builder.get_object(
            'filter_btn'
        )
        self.filter_popover = FeedsViewPopover(self.filter_btn)
        self.filter_btn.set_popover(self.filter_popover)
        # this activates the "All" feed filter. while this works it's kinda
        # hacky and needs a proper function
        self.feedman.connect(
            'feedmanager_refresh_start',
            lambda caller, msg:
            self.filter_popover.scrolled_win.listbox.row_all_activate(
                skip=(msg == 'startup')
            )
        )

        self.add_btn = self.builder.get_object(
            'add_btn'
        )
        self.add_btn.set_tooltip_text(_('Add new feed'))
        self.add_popover = AddFeedPopover(self.add_btn)
        self.add_btn.set_popover(self.add_popover)

        self.refresh_btn = RefreshSpinnerButton()
        self.refresh_btn.btn.connect('clicked', self.feedman.refresh)
        self.builder.get_object('refresh_btn_box').add(self.refresh_btn)

        self.squeezer = Handy.Squeezer(orientation=Gtk.Orientation.HORIZONTAL)
        self.squeezer.set_homogeneous(False)
        self.squeezer.set_interpolate_size(False)
        self.squeezer.set_hexpand(False)
        self.nobox = Gtk.Label()
        self.nobox.set_size_request(1, -1)
        self.stack_switcher = Handy.ViewSwitcher()
        self.stack_switcher.set_policy(Handy.ViewSwitcherPolicy.WIDE)
        self.stack_switcher.set_margin_start(12)
        self.stack_switcher.set_margin_end(12)
        self.squeezer.add(self.stack_switcher)
        self.squeezer.add(self.nobox)
        self.squeezer.connect('notify::visible-child', self.on_squeeze)
        self.left_headerbar.set_custom_title(self.squeezer)

        self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        self.feedman.connect(
            'feedmanager_refresh_start',
            self.on_new_feed_add_start
        )
        self.feedman.connect(
            'feedmanager_refresh_end',
            self.on_new_feed_add_end
        )
        self.title_squeezer = self.builder.get_object(
            'right_headerbar_squeezer'
        )
        self.right_title_container = self.builder.get_object(
            'right_headerbar_title_container'
        )
        self.title_label = self.builder.get_object('title_label')
        self.title_squeezer.add(self.right_title_container)
        self.title_squeezer.add(Gtk.Label(label=''))

    def set_view_mode_icon(self, mode):
        self.view_mode_menu_btn_icon.set_from_icon_name(
            {
                'webview': 'globe-alt-symbolic',
                'reader': 'ephy-reader-mode-symbolic',
                'rsscont': 'application-rss+xml-symbolic'
            }[mode],
            Gtk.IconSize.BUTTON
        )

    def on_view_mode_change(self, target):
        self.view_mode_menu.popdown()
        if target == 'webview':
            self.webview.set_enable_reader_mode(False)
        elif target == 'reader':
            self.webview.set_enable_reader_mode(True)
        elif target == 'rsscont':
            self.webview.set_enable_rss_content(True)
        self.set_view_mode_icon(target)

    def set_article_title(self, title):
        self.title_label.set_text(title)

    def on_squeeze(self, *args):
        self.emit(
            'gfeeds_headerbar_squeeze',
            self.squeezer.get_visible_child() == self.nobox
        )

    def on_search_btn_toggled(self, togglebtn):
        self.searchbar.set_search_mode(togglebtn.get_active())

    def on_new_feed_add_start(self, *args):
        self.refresh_btn.set_spinning(True)
        self.add_popover.confirm_btn.set_sensitive(False)

    def on_new_feed_add_end(self, *args):
        self.refresh_btn.set_spinning(False)
        self.add_popover.confirm_btn.set_sensitive(True)
        self.add_popover.url_entry.set_text('')

    def set_headerbar_controls(self, *args):
        self.right_headerbar.set_show_close_button(True)
        self.left_headerbar.set_show_close_button(True)

    def on_back_button_clicked(self, *args):
        self.leaflet.set_visible_child(self.left_headerbar)
        self.back_btn_func()

    def on_load_start(self, *args):
        self.view_mode_menu_btn.set_sensitive(False)

    def on_load_end(self, *args):
        self.view_mode_menu_btn.set_sensitive(True)

    def copy_article_uri(self, *args):
        self.clipboard.set_text(self.webview.uri, -1)
        self.clipboard.store()
        self.webview.show_notif()



from gettext import gettext as _
from gi.repository import Gtk


def GFeedsOpmlFileChooserDialog(parent_window):
    dialog = Gtk.FileChooserNative.new(
        _('Choose an OPML file to import'),
        parent_window,
        Gtk.FileChooserAction.OPEN,
        Gtk.STOCK_OPEN,
        Gtk.STOCK_CANCEL
    )
    f_filter = Gtk.FileFilter()
    f_filter.set_name(_('XML files'))
    f_filter.add_mime_type('text/xml')
    dialog.add_filter(f_filter)
    return dialog


def GFeedsOpmlSavePathChooserDialog(parent_window):
    dialog = Gtk.FileChooserNative.new(
        _('Choose where to save the exported OPML file'),
        parent_window,
        Gtk.FileChooserAction.SAVE,
        Gtk.STOCK_SAVE,
        Gtk.STOCK_CANCEL
    )
    dialog.set_do_overwrite_confirmation(True)
    dialog.set_create_folders(True)
    f_filter = Gtk.FileFilter()
    f_filter.set_name(_('XML files'))
    f_filter.add_mime_type('text/xml')
    dialog.add_filter(f_filter)
    dialog.set_current_name('GFeeds.opml')
    return dialog
